"""聊天插件API模块

该模块提供了聊天插件的公共API接口，包括菜单管理、管理员操作、聊天功能等核心功能的封装。
"""

from __future__ import annotations

from nonebot import logger

from amrita.utils.admin import send_to_admin

from .config import Config, ConfigManager, config_manager
from .on_event import on_before_chat, on_before_poke, on_chat, on_event, on_poke
from .utils.libchat import (
    AdapterManager,
    ModelAdapter,
    get_chat,
    tools_caller,
)
from .utils.llm_tools.manager import ToolsManager, on_tools
from .utils.llm_tools.mcp_client import MCP_SERVER_SCRIPT_TYPE, ClientManager, MCPClient
from .utils.llm_tools.models import (
    FunctionDefinitionSchema,
    FunctionParametersSchema,
    FunctionPropertySchema,
    ToolChoice,
    ToolContext,
    ToolData,
    ToolFunctionSchema,
)
from .utils.memory import get_memory_data
from .utils.models import InsightsModel
from .utils.tokenizer import Tokenizer, hybrid_token_count


class Admin:
    """管理员管理类

    负责处理与管理员相关的操作，如发送消息、错误处理和管理员权限管理。
    """

    config: Config

    def __init__(self):
        """构造函数，初始化配置"""
        self.config = config_manager.ins_config

    async def send_with(self, msg: str) -> Admin:
        """异步发送消息给管理员。

        参数:
        - msg (str): 要发送的消息内容。

        返回:
        - Admin: 返回Admin实例，支持链式调用。
        """
        await send_to_admin(msg)
        return self

    async def send_error(self, msg: str) -> Admin:
        """异步发送错误消息给管理员，并记录错误日志。

        参数:
        - msg (str): 要发送的错误消息内容。

        返回:
        - Admin: 返回Admin实例，支持链式调用。
        """
        logger.error(msg)
        await send_to_admin(msg)
        return self


class Chat:
    """聊天处理类

    Chat 类用于处理与LLM相关操作，如获取消息响应、调用工具等。
    """

    config: Config

    def __init__(self):
        """构造函数，初始化配置"""
        self.config = config_manager.ins_config

    async def get_msg(self, prompt: str, message: list):
        """获取LLM响应

        在消息列表前插入系统提示词，然后调用模型获取响应。

        :param prompt[str]: 系统提示词
        :param message[list]: 消息列表

        :returns: 模型响应结果
        """
        message.insert(0, {"role": "assistant", "content": prompt})
        return await self.get_msg_on_list(message)

    async def get_msg_on_list(self, message: list):
        """获取LLM响应

        直接使用提供的消息列表调用模型获取响应。

        :param message[list]: 消息列表

        :returns: 模型响应结果
        """
        return await get_chat(messages=message)

    async def call_tools(
        self,
        messages: list,
        tools: list,
        tool_choice: ToolChoice | None = None,
    ):
        """调用工具

        使用指定的工具和消息调用工具函数。

        :param messages: 消息列表
        :param tools: 工具列表
        :param tool_choice: 工具选择参数（可选）

        :returns: 工具调用结果
        """
        return await tools_caller(
            messages=messages, tools=tools, tool_choice=tool_choice
        )


__all__ = [
    "MCP_SERVER_SCRIPT_TYPE",
    "AdapterManager",
    "Admin",
    "Chat",
    "ClientManager",
    "ConfigManager",
    "FunctionDefinitionSchema",
    "FunctionParametersSchema",
    "FunctionPropertySchema",
    "InsightsModel",
    "MCPClient",
    "ModelAdapter",
    "Tokenizer",
    "ToolContext",
    "ToolData",
    "ToolFunctionSchema",
    "ToolsManager",
    "config_manager",
    "get_memory_data",
    "hybrid_token_count",
    "on_before_chat",
    "on_before_poke",
    "on_chat",
    "on_event",
    "on_poke",
    "on_tools",
    "tools_caller",
]



from nonebot import get_driver, logger

from amrita.plugins.chat.utils.llm_tools.mcp_client import ClientManager

from . import config
from .config import config_manager
from .hook_manager import run_hooks

driver = get_driver()
__LOGO = "\033[34mLoading SuggarChat \033[33m {version}-Amrita......\033[0m"


@driver.on_startup
async def onEnable():
    kernel_version = "V3"
    config.__kernel_version__ = kernel_version
    logger.info(__LOGO.format(version=kernel_version))
    logger.debug("加载配置文件...")
    await config_manager.safe_get_config()
    await run_hooks()
    if (conf := config.config_manager.config).llm_config.tools.agent_mcp_client_enable:
        logger.info("正在初始化MCP Client......")
        mcp_servers = conf.llm_config.tools.agent_mcp_server_scripts
        for server in mcp_servers:
            try:
                await ClientManager().initialize_this(server)
            except Exception as e:  # noqa: PERF203
                logger.error(f"初始化MCP Client@{server}失败: {e}")
                logger.opt(exception=e, colors=True).exception(e)
        logger.info("MCP Client初始化完成！")
    logger.debug("成功启动！")



from nonebot.adapters import Message
from nonebot.adapters.onebot.v11.event import MessageEvent
from nonebot.matcher import Matcher
from nonebot.params import CommandArg

from ..config import config_manager


async def set_preset(
    event: MessageEvent, matcher: Matcher, args: Message = CommandArg()
):
    """处理设置预设的事件"""

    # 检查插件是否启用
    if not config_manager.ins_config.enable:
        matcher.skip()

    # 获取命令参数并去除多余空格
    arg = args.extract_plain_text().strip()

    # 如果参数不为空
    if arg != "":
        # 遍历所有模型
        for model in await config_manager.get_all_presets():
            if model.name == arg:
                # 设置预设并保存
                config_manager.ins_config.preset = model.name
                await config_manager.save_config()
                # 回复设置成功
                await matcher.finish(f"已设置预设为：{model.name}，模型：{model.model}")
        # 未找到匹配的预设
        await matcher.finish("未找到预设，请输入/presets查看预设列表。")
    else:
        # 参数为空时重置为默认预设
        config_manager.ins_config.preset = "default"
        await config_manager.save_config()
        # 回复重置成功
        await matcher.finish(
            f"已重置预设为：默认预设，模型：{(await config_manager.get_preset(config_manager.ins_config.preset)).model}。"
        )



import copy
import json
import os
import re
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, ClassVar, Literal, TypeVar

import aiofiles
import nonebot_plugin_localstore as store
import tomli
import tomli_w
from nonebot import get_driver, logger
from pydantic import BaseModel, Field

from amrita.config_manager import BaseDataManager, UniConfigManager

__kernel_version__ = "unknown"

# 保留为其他插件提供的引用

# 配置目录
CONFIG_DIR: Path = store.get_plugin_config_dir()
driver = get_driver()
nb_config = driver.config
STRDICT = dict[str, Any]

T = TypeVar("T", STRDICT, list[str | STRDICT], str)

# 缓存的正则表达式
_re_hash: int = 0
_cached_pattern: re.Pattern[str] | None = None


def replace_env_vars(
    data: T,
) -> T:
    """递归替换环境变量占位符，但不修改原始数据"""
    data_copy = copy.deepcopy(data)  # 创建原始数据的深拷贝[4,5](@ref)
    if isinstance(data_copy, dict):
        for key, value in data_copy.items():
            data_copy[key] = replace_env_vars(value)
    elif isinstance(data_copy, list):
        for i in range(len(data_copy)):
            data_copy[i] = replace_env_vars(data_copy[i])
    elif isinstance(data_copy, str):
        patterns = (
            r"\$\{(\w+)\}",
            r"\{\{(\w+)\}\}",
        )  # 支持两种格式的占位符，分别为 ${} 和 {{}}

        def replacer(match: re.Match[str]) -> str:
            var_name = match.group(1)
            return os.getenv(var_name, "")  # 若未设置环境变量，返回空字符串

        for pattern in patterns:
            if re.search(pattern, data_copy):
                # 如果匹配到占位符，则进行替换
                data_copy = re.sub(pattern, replacer, data_copy)
                break  # 替换后跳出循环，避免重复替换
    return data_copy


class ModelPreset(BaseModel):
    model: str = Field(default="", description="使用的AI模型名称（如gpt-3.5-turbo）")
    name: str = Field(default="default", description="当前预设的标识名称")
    base_url: str = Field(
        default="", description="API服务的基础地址（为空则使用OpenAI默认地址）"
    )
    api_key: str = Field(default="", description="访问API所需的密钥")
    protocol: str = Field(default="__main__", description="协议适配器类型")
    thought_chain_model: bool = Field(
        default=False, description="是否启用思维链模型优化（增强复杂问题处理）"
    )
    multimodal: bool = Field(
        default=False, description="是否支持多模态输入（如图片识别）"
    )
    extra: dict[str, Any] = Field(default_factory=dict)

    @classmethod
    def load(cls, path: Path):
        if path.exists():
            with path.open(
                "r",
                encoding="utf-8",
            ) as f:
                data = json.load(f)
            return cls.model_validate(data)
        return cls()  # 返回默认值

    def save(self, path: Path):
        with path.open("w", encoding="u8") as f:
            json.dump(self.model_dump(), f, indent=4, ensure_ascii=False)


class ToolsConfig(BaseModel):
    enable_tools: bool = Field(
        default=True,
        description="是否启用外部工具调用功能（关闭此选项不影响内容审查系统）",
    )
    use_minimal_context: bool = Field(
        default=True,
        description="是否使用最小上下文，即使用系统prompt+用户最后一条消息（关闭此选项将使用消息列表的所有上下文，在Agent工作流执行中可能会消耗大量Tokens，启用此选项可能会有效降低Tokens使用量）",
    )
    enable_report: bool = Field(default=True, description="是否启用内容审查系统")
    report_exclude_system_prompt: bool = Field(
        default=False,
        description="是否排除系统提示词，默认情况下，内容审查会检查系统提示和上下文",
    )
    report_exclude_context: bool = Field(
        default=False,
        description="是否排除上下文，默认情况下，内容审查会检查系统提示和上下文",
    )
    report_then_block: bool = Field(
        default=True, description="检测到违规内容后是否熔断会话"
    )
    require_tools: bool = Field(
        default=False, description="是否强制要求每次调用至少使用一个工具"
    )
    agent_mode_enable: bool = Field(
        default=False, description="启用智能体模式（实验体验）"
    )
    agent_tool_call_limit: int = Field(
        default=10, description="智能体模式下的工具调用限制"
    )
    agent_tool_call_notice: Literal["hide", "notify"] = Field(
        default="hide",
        description="智能体模式下的工具调用情况提示方式，hide为隐藏，notify为通知",
    )
    agent_thought_mode: Literal[
        "reasoning", "chat", "reasoning-required", "reasoning-optional"
    ] = Field(
        default="chat",
        description="智能体模式下的思考模式，reasoning模式会先执行思考过程，然后执行任务；"
        "reasoning-required要求每次Tool Calling都执行任务分析；"
        "reasoning-optional不要求reasoning，但是允许reasoning；"
        "chat模式会直接执行任务",
    )
    agent_reasoning_hide: bool = Field(
        default=False, description="是否隐藏智能体模式下的思考过程"
    )
    agent_mcp_client_enable: bool = Field(
        default=False, description="是否启用MCP客户端"
    )
    agent_mcp_server_scripts: list[str] = Field(
        default=[], description="MCP服务端脚本列表"
    )


class SessionConfig(BaseModel):
    session_control: bool = Field(default=False, description="是否启用会话超时自动清理")
    session_control_time: int = Field(
        default=60, description="会话超时时间（单位：分钟）"
    )
    session_control_history: int = Field(
        default=10, description="会话历史记录最大保存条数"
    )
    session_max_tokens: int = Field(
        default=5000, description="单次会话上下文最大token容量"
    )


class AutoReplyConfig(BaseModel):
    enable: bool = Field(default=False, description="是否启用自动回复系统")
    global_enable: bool = Field(
        default=False, description="是否全局启用自动回复（无视会话状态）"
    )
    probability: float = Field(default=1e-2, description="随机触发概率（0.01=1%）")
    keywords: list[str] = Field(default=["at"], description="触发自动回复的关键字列表")
    keywords_mode: Literal["starts_with", "contains"] = Field(
        default="starts_with", description="自动回复配置(starts_with/contains)"
    )


class FunctionConfig(BaseModel):
    chat_pending_mode: Literal["single", "queue", "single_with_report"] = Field(
        default="queue",
        description="聊天时，如果同一个Session并发调用但是上一条消息没有处理完时插件的行为。\n"
        + "single: 忽略这条消息；\n"
        + "queue: 等待上一条消息处理完再处理；\n"
        + "single_with_report: 忽略这条消息并提示用户正在等待。",
    )
    synthesize_forward_message: bool = Field(
        default=True, description="是否解析合并转发消息"
    )
    nature_chat_style: bool = Field(
        default=True, description="是否启用自然对话风格优化(自动分句)"
    )
    nature_chat_cut_pattern: str = Field(
        default=r'([。！？!?;；\n]+)[""\'\'"\s]*', description="分句功能的正则表达式"
    )
    poke_reply: bool = Field(default=True, description="是否响应戳一戳事件")
    enable_group_chat: bool = Field(default=True, description="是否启用群聊功能")
    enable_private_chat: bool = Field(default=True, description="是否启用私聊功能")
    allow_custom_prompt: bool = Field(
        default=True, description="是否允许用户自定义提示词"
    )
    use_user_nickname: bool = Field(
        default=False, description="在群聊中使用QQ昵称而非群名片"
    )

    @property
    def pattern(self) -> re.Pattern:
        """
        获取分句的正则表达式
        """
        global _cached_pattern, _re_hash
        pattern_hash = hash(self.nature_chat_cut_pattern)
        if pattern_hash != _re_hash or _cached_pattern is None:
            _cached_pattern = re.compile(self.nature_chat_cut_pattern)
            _re_hash = pattern_hash
        return _cached_pattern


class PresetSwitch(BaseModel):
    backup_preset_list: list[str] = Field(
        default=[], description="主模型不可用时自动切换的备选模型预设列表"
    )
    multi_modal_preset_list: list[str] = Field(
        default=[], description="多模态场景预设调用顺序"
    )


class CookieModel(BaseModel):
    cookie: str = Field(default="", description="用于安全检测的Cookie字符串")
    enable_cookie: bool = Field(default=False, description="是否启用Cookie泄露检测机制")

    @property
    def block_msg(self) -> list[str]:
        return ConfigManager().config.llm_config.block_msg

    @block_msg.setter
    def block_msg(self, value: list[str]):
        ConfigManager().config.llm_config.block_msg = value


class ExtendConfig(BaseModel):
    say_after_self_msg_be_deleted: bool = Field(
        default=False, description="消息被撤回后是否自动回复"
    )
    group_added_msg: str = Field(
        default="你好，我是Suggar，欢迎使用SuggarAI聊天机器人...",
        description="入群欢迎消息",
    )
    send_msg_after_be_invited: bool = Field(
        default=False, description="被邀请入群后是否主动发言"
    )
    after_deleted_say_what: list[str] = Field(
        default=[
            "抱歉啦，不小心说错啦～",
            "嘿，发生什么事啦？我",
            "唔，我是不是说错了什么？",
            "纠错时间到，如果我说错了请告诉我！",
            "发生了什么？我刚刚没听清楚呢~",
            "我会记住的，绝对不再说错话啦~",
            "哦，看来我又犯错了，真是不好意思！",
            "哈哈，看来我得多读书了~",
            "哎呀，真是个小口误，别在意哦~",
            "哎呀，我也有尴尬的时候呢~",
            "希望我能继续为你提供帮助，不要太在意我的小错误哦！",
        ],
        description="消息被撤回后的随机回复列表",
    )


class UsageLimitConfig(BaseModel):
    enable_usage_limit: bool = Field(default=False, description="是否启用使用频率限制")
    group_daily_limit: int = Field(default=100, description="单个群组每日最大使用次数")
    user_daily_limit: int = Field(default=100, description="单个用户每日最大使用次数")
    group_daily_token_limit: int = Field(
        default=200000, description="单个群组每日最大token消耗量"
    )
    user_daily_token_limit: int = Field(
        default=100000, description="单个用户每日最大token消耗量"
    )
    total_daily_limit: int = Field(default=1500, description="总使用次数限制")
    total_daily_token_limit: int = Field(default=1000000, description="总使用token限制")
    global_insights_expire_days: int = Field(default=7, description="全局统计过期天数")


class LLM_Config(BaseModel):
    tools: ToolsConfig = Field(default=ToolsConfig(), description="工具调用配置")
    stream: bool = Field(default=False, description="是否启用流式响应（逐字输出）")
    memory_lenth_limit: int = Field(default=50, description="记忆上下文的最大消息数量")
    max_tokens: int = Field(default=100, description="单次回复生成的最大token数")
    tokens_count_mode: Literal["word", "bpe", "char"] = Field(
        default="bpe", description="Token计算模式：bpe(子词)/word(词语)/char(字符)"
    )
    enable_tokens_limit: bool = Field(
        default=True, description="是否启用上下文长度限制"
    )
    llm_timeout: int = Field(default=60, description="API请求超时时间（秒）")
    auto_retry: bool = Field(default=True, description="请求失败时自动重试")
    max_retries: int = Field(default=3, description="最大重试次数")
    enable_memory_abstract: bool = Field(
        default=True,
        description="是否启用上下文记忆摘要(将删除上下文替换为一个摘要插入到system instruction中)",
    )
    memory_abstract_proportion: float = Field(
        default=15e-2, description="上下文摘要比例(0.15=15%)"
    )
    block_msg: list[str] = Field(
        default=["你好，这个问题我暂时无法处理，请稍后再试。"],
        description="触发安全熔断时随机返回的提示消息",
    )


class Config(BaseModel):
    preset_extension: PresetSwitch = Field(
        default=PresetSwitch(), description="预设模型扩展配置"
    )
    default_preset: ModelPreset = Field(
        default=ModelPreset(), description="默认预设配置"
    )
    session: SessionConfig = Field(default=SessionConfig(), description="会话管理配置")
    cookies: CookieModel = Field(
        default=CookieModel(), description="电子水印检测功能配置"
    )
    autoreply: AutoReplyConfig = Field(
        default=AutoReplyConfig(), description="自动回复设置"
    )
    function: FunctionConfig = Field(
        default=FunctionConfig(), description="功能开关配置"
    )
    extended: ExtendConfig = Field(default=ExtendConfig(), description="扩展行为设置")
    llm_config: LLM_Config = Field(default=LLM_Config(), description="大语言模型配置")
    extra: dict[str, Any] = Field(default={}, description="扩展预留区")
    usage_limit: UsageLimitConfig = Field(
        default=UsageLimitConfig(), description="使用限额配置"
    )
    enable: bool = Field(default=False, description="是否启用 SuggarChat 主功能")
    parse_segments: bool = Field(
        default=True, description="是否解析特殊消息段（如@提及/合并转发等）"
    )
    matcher_function: bool = Field(
        default=True, description="是否启用 SuggarMatcher 高级匹配功能"
    )
    preset: str = Field(default="default", description="默认使用的模型预设配置名称")
    group_prompt_character: str = Field(
        default="default", description="群聊场景使用的提示词模板名称"
    )
    private_prompt_character: str = Field(
        default="default", description="私聊场景使用的提示词模板名称"
    )

    @classmethod
    def load_from_toml(cls, path: Path) -> "Config":
        """从 TOML 文件加载配置"""
        if not path.exists():
            return cls()
        with open(str(path), encoding="u8") as f:
            data: dict[str, Any] = tomli.loads(f.read())
        return cls.model_validate(data)

    def validate_value(self):
        """校验配置"""
        if self.llm_config.max_tokens <= 0:
            raise ValueError("max_tokens必须大于零!")
        if self.llm_config.llm_timeout <= 0:
            raise ValueError("LLM请求超时时间必须大于零！")
        if self.session.session_max_tokens <= 0:
            raise ValueError("上下文最大Tokens限制必须大于零！")
        if self.session.session_control:
            if self.session.session_control_history <= 0:
                raise ValueError("会话历史最大值不能为0！")
            if self.session.session_control_time <= 0:
                raise ValueError("会话生命周期时间不能小于零！")

    @classmethod
    def load_from_json(cls, path: Path) -> "Config":
        """从 JSON 文件加载配置"""
        with path.open("r", encoding="utf-8") as f:
            data: dict[str, Any] = json.load(f)
        return cls.model_validate(data)

    def save_to_toml(self, path: Path):
        """保存配置到 TOML 文件"""
        with path.open("w", encoding="utf-8") as f:
            f.write(tomli_w.dumps(self.model_dump()))


@dataclass
class Prompt:
    text: str = ""
    name: str = "default"


@dataclass
class Prompts:
    group: list[Prompt] = field(default_factory=list)
    private: list[Prompt] = field(default_factory=list)

    def save_group(self, path: Path):
        """保存群组提示词"""
        for prompt in self.group:
            with (path / f"{prompt.name}.txt").open(
                "w",
                encoding="u8",
            ) as f:
                f.write(prompt.text)

    def save_private(self, path: Path):
        """保存私聊提示词"""
        for prompt in self.private:
            with (path / f"{prompt.name}.txt").open(
                "w",
                encoding="u8",
            ) as f:
                f.write(prompt.text)


class ConfigManager(BaseDataManager[Config]):
    config_dir: Path = CONFIG_DIR
    private_prompts: Path = config_dir / "private_prompts"
    group_prompts: Path = config_dir / "group_prompts"
    custom_models_dir: Path = config_dir / "models"
    _private_train: ClassVar[dict[str, Any]] = {}
    _group_train: ClassVar[dict[str, Any]] = {}
    _model_name2file: ClassVar[dict[str, Path]] = {}
    ins_config: Config = Config()
    models: ClassVar[list[tuple[ModelPreset, str]]] = []
    prompts: Prompts = Prompts()
    config: Config
    _config_id: int | None = None
    _cached_env_config: Config | None = None
    _owner_name = store._try_get_caller_plugin().name
    __lateinit__ = True

    def __getattribute__(self, name: str) -> Any:
        if name == "config":
            conf_id = id(self.ins_config)
            if conf_id == self._config_id:
                assert self._cached_env_config
                return self._cached_env_config
            self._config_id = conf_id
            conf_data: dict[str, Any] = self.ins_config.model_dump()
            result = replace_env_vars(conf_data)
            self._cached_env_config = Config.model_validate(result)
            return self._cached_env_config
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "config":
            self._cached_env_config = None
            self._config_id = None
            self.ins_config = value
        return super().__setattr__(name, value)

    async def __apost_init__(self):
        return await self.load()

    async def load(self):
        """_初始化配置目录_"""

        async def prompt_callback():
            logger.info("正在重载插件提示词文件...")
            await self.get_prompts(False, True)
            await self.load_prompt()
            logger.success("提示词文件已重载")

        async def models_callback():
            logger.info("正在重载模型目录...")
            await self.get_all_presets(False)
            logger.success("完成")

        logger.info("正在初始化存储目录...")
        logger.debug(f"配置目录: {self.config_dir}")
        os.makedirs(self.config_dir, exist_ok=True)
        os.makedirs(self.private_prompts, exist_ok=True)
        os.makedirs(self.group_prompts, exist_ok=True)
        os.makedirs(self.custom_models_dir, exist_ok=True)

        await UniConfigManager().add_directory("models", lambda *_: models_callback())
        self.validate_presets()
        ps = await self.get_all_presets(cache=False)
        logger.info(f"加载了{len(ps)}个模型")
        p = await self.get_prompts(cache=False)
        logger.info(f"加载了{len(p.group) + len(p.private)}个提示词")
        await self.load_prompt()
        await UniConfigManager().add_directory(
            "group_prompts",
            lambda *_: prompt_callback(),
            lambda change: (change[1].startswith(str(self.group_prompts)))
            and change[1].endswith(".txt"),
        )
        await UniConfigManager().add_directory(
            "private_prompts",
            lambda *_: prompt_callback(),
            lambda change: change[1].startswith(str(self.private_prompts))
            and change[1].endswith(".txt"),
        )

    def validate_presets(self):
        def validate_preset(path: Path):
            try:
                model_data = ModelPreset.load(path)
                model_data.save(path)
                self._model_name2file[model_data.name] = path
            except Exception as e:
                logger.opt(colors=True).error(
                    f"Failed to validate preset '{file!s}' because '{e!s}'"
                )

        for file in self.custom_models_dir.glob("*.json"):
            validate_preset(file)

    async def get_all_presets(self, cache: bool = False) -> list[ModelPreset]:
        """获取模型列表"""
        if cache and self.models:
            return [model for model, _ in self.models]
        self.models.clear()  # 清空模型列表

        for file in self.custom_models_dir.glob("*.json"):
            model_data = ModelPreset.load(file).model_dump()
            preset_data = replace_env_vars(model_data)
            if not isinstance(preset_data, dict):
                raise TypeError("Expected replace_env_vars to return a dict")
            model_preset = ModelPreset.model_validate(preset_data)
            self._model_name2file[model_preset.name] = file
            self.models.append((model_preset, file.stem))

        return [model for model, _ in self.models]

    async def get_preset(
        self, preset: str, fix: bool = False, cache: bool = False
    ) -> ModelPreset:
        """_获取预设配置_

        Args:
            preset (str): _预设的字符串名称_
            fix (bool, optional): _是否修正不存在的预设_. Defaults to False.
            cache (bool, optional): _是否使用缓存_. Defaults to False.

        Returns:
            ModelPreset: _模型预设对象_
        """
        if preset == "default":
            return config_manager.config.default_preset
        for model in await self.get_all_presets(cache=cache):
            if model.name == preset:
                return model
        if fix:
            config_manager.ins_config.preset = "default"
            await config_manager.save_config()
        return await self.get_preset("default", fix, cache)

    async def get_prompts(
        self, cache: bool = False, load_only: bool = False
    ) -> Prompts:
        """获取提示词"""
        if cache and self.prompts:
            return self.prompts
        self.prompts = Prompts()
        for file in self.private_prompts.glob("*.txt"):
            async with aiofiles.open(file, encoding="utf-8") as f:
                prompt = await f.read()
            self.prompts.private.append(Prompt(prompt, file.stem))
        for file in self.group_prompts.glob("*.txt"):
            async with aiofiles.open(file, encoding="utf-8") as f:
                prompt = await f.read()
            self.prompts.group.append(Prompt(prompt, file.stem))
        if not self.prompts.private:
            self.prompts.private.append(Prompt("", "default"))
        if not self.prompts.group:
            self.prompts.group.append(Prompt("", "default"))

        if not load_only:
            self.prompts.save_private(self.private_prompts)
            self.prompts.save_group(self.group_prompts)

        return self.prompts

    @property
    def private_train(self) -> dict[str, str]:
        """获取私聊提示词"""
        return deepcopy(self._private_train)

    @property
    def group_train(self) -> dict[str, str]:
        """获取群聊提示词"""
        return deepcopy(self._group_train)

    async def load_prompt(self):
        """加载提示词，匹配预设"""
        for prompt in self.prompts.group:
            if prompt.name == self.ins_config.group_prompt_character:
                self.__class__._group_train = {"role": "system", "content": prompt.text}
                break
        else:
            self.__class__._group_train = {
                "role": "system",
                "content": next(
                    i for i in self.prompts.group if i.name == "default"
                ).text,
            }
            logger.warning(
                f"没有找到名称为 {self.ins_config.group_prompt_character} 的群组提示词，将使用default.txt!"
            )

        for prompt in self.prompts.private:
            if prompt.name == self.__class__.ins_config.private_prompt_character:
                self.__class__._private_train = {
                    "role": "system",
                    "content": prompt.text,
                }
                break
        else:
            logger.warning(
                f"没有找到名称为 {self.__class__.ins_config.private_prompt_character} 的私聊提示词，将使用default.txt！"
            )
            self.__class__._private_train = {
                "role": "system",
                "content": next(
                    i for i in self.prompts.private if i.name == "default"
                ).text,
            }

    async def save_config(self):
        """保存配置"""
        await UniConfigManager().save_config(self._owner_name)

    async def set_config(self, key: str, value: str):
        """
        设置配置

        :param key: 配置项的名称
        :param value: 配置项的值

        :raises KeyError: 如果配置项不存在，则抛出异常
        """
        if hasattr(self.ins_config, key):
            setattr(self.ins_config, key, value)
            await self.save_config()
        else:
            raise KeyError(f"配置项 {key} 不存在")

    async def register_config(self, key: str, default_value=None):
        """
        注册配置项

        :param key: 配置项的名称

        """
        if default_value is None:
            default_value = "null"
        self.ins_config.extra.setdefault(key, default_value)
        await self.save_config()

    def reg_config(self, key: str, default_value=None):
        """
        注册配置项

        :param key: 配置项的名称

        """
        return self.register_config(key, default_value)

    def reg_model_config(self, key: str, default_value=None):
        """
        注册模型配置项

        :param key: 配置项的名称

        """
        if default_value is None:
            default_value = "null"
        if key not in self.ins_config.default_preset.extra:
            self.ins_config.default_preset.extra.setdefault(key, default_value)
        for model, name in self.models:
            model.extra.setdefault(key, default_value)
            model.save(self.custom_models_dir / f"{name}.json")


config_manager = ConfigManager()



import importlib
import os
from typing import Any

import nonebot
import nonebot.plugin.manager
from nonebot.plugin.model import Plugin
from nonebot.utils import escape_tag

import amrita

logger = nonebot.logger


class LoadError(Exception):
    def __init__(self, data: Any):
        self.data = data


def _load_plugin(self, name: str) -> Plugin | None:
    """加载指定插件。

    可以使用完整插件模块名或者插件标识符加载。

    参数:
        name: 插件名称或插件标识符。
    """

    try:
        # load using plugin id
        if name in self._third_party_plugin_ids:
            module = importlib.import_module(self._third_party_plugin_ids[name])
        elif name in self._searched_plugin_ids:
            module = importlib.import_module(self._searched_plugin_ids[name])
        # load using module name
        elif (
            name in self._third_party_plugin_ids.values()
            or name in self._searched_plugin_ids.values()
        ):
            module = importlib.import_module(name)
        else:
            raise RuntimeError(f"Plugin not found: {name}! Check your plugin name")

        if (plugin := getattr(module, "__plugin__", None)) is None or not isinstance(
            plugin, Plugin
        ):
            raise RuntimeError(
                f"Module {module.__name__} is not loaded as a plugin! "
                f"Make sure not to import it before loading."
            )
        logger.opt(colors=True).success(
            f'Succeeded to load plugin "<y>{escape_tag(plugin.id_)}</y>"'
            + (
                f' from "<m>{escape_tag(plugin.module_name)}</m>"'
                if plugin.module_name != plugin.id_
                else ""
            )
        )
        return plugin
    except Exception as e:
        logger.opt(colors=True, exception=e).error(
            f'<r><bg #f8bbd0>Failed to import "{escape_tag(name)}"</bg #f8bbd0></r>'
        )
        raise LoadError(e) from e


def main():
    global __name__
    __name__ = "Amrita Loader"
    nonebot.plugin.manager.PluginManager.load_plugin = _load_plugin

    os.environ["LOG_LEVEL"] = "DEBUG"
    os.environ["ALEMBIC_STARTUP_CHECK"] = "false"
    print("---AMRITA LOADER IS LOADING---")
    amrita.init()
    logger.info("Loading driver...")
    driver = nonebot.get_driver()
    logger.info("Loading plugins...")
    try:
        amrita.load_plugins()
        logger.info("plugin import done!")
    except LoadError:
        logger.error("Test FAILED! Good luck to you next time!")
        exit(1)
    except Exception as e:
        logger.error("OOPS!There is something wrong while this process is running!")
        logger.opt(exception=True).error(f"Error!：{type(e).__name__}")
        exit(1)
    driver.on_startup(exit_test)
    logger.info("Testing pre-startup...")
    amrita.run()


async def exit_test():
    logger.info("Finished!")
    os._exit(0)



"""memory_abstract

迁移 ID: 1e67057a1646
父迁移: 8b619d7fca40
创建时间: 2026-01-02 18:53:23.916959

"""

from __future__ import annotations

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

revision: str = "1e67057a1646"
down_revision: str | Sequence[str] | None = "8b619d7fca40"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("suggarchat_memory_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "memory_abstract",
                sa.Text(),
                nullable=False,
                server_default="",
            )
        )

    # ### end Alembic commands ###


def downgrade(name: str = "") -> None:
    if name:
        return
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("suggarchat_memory_data", schema=None) as batch_op:
        batch_op.drop_column(
            "memory_abstract",
        )

    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
import random
import re
import secrets
from abc import ABC
from asyncio import Lock
from datetime import datetime, timedelta
from typing import TypeVar, overload

import bcrypt
from fastapi import HTTPException, Request
from nonebot import logger
from pydantic import BaseModel, Field
from pytz import utc
from typing_extensions import Self

from amrita.plugins.webui.service.config import get_webui_config

T = TypeVar("T")
BOT_SESSION_ID = random.randint(0, 1000000)
TOKEN_KEY = f"amrita_token_{BOT_SESSION_ID}"


def get_restful_auth_header(request: Request) -> str | None:
    if auth_header := request.headers.get("Authorization"):
        search = r"Bearer (.+)"
        result = re.search(search, auth_header)
        if not result:
            raise HTTPException(status_code=403, detail="未授权的访问")
        return result.group(1)


class NOT_GIVEN(ABC):
    @classmethod
    def __bool__(cls):
        return False


class TokenData(BaseModel):
    username: str
    expire: datetime = Field(
        default_factory=lambda: datetime.now(utc) + timedelta(minutes=30)
    )
    onetime_token: str | None = None


class OnetimeTokenData(BaseModel):
    token: str
    expire: datetime = Field(
        default_factory=lambda: datetime.now(utc) + timedelta(minutes=10)
    )


class LoginRateLimiter:
    """
    登录速率限制器，防止暴力破解
    """

    _instance = None
    __limiter_lock: Lock
    # 存储IP地址和对应的请求时间列表
    __requests: dict[str, list[datetime]]

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls.__limiter_lock = Lock()
            cls.__requests = {}
        return cls._instance

    async def is_allowed(self, ip: str, count: int = 10) -> bool:
        """
        检查指定IP是否允许登录请求
        1秒内最多允许10次请求
        """
        async with self.__limiter_lock:
            now = datetime.now(utc)
            # 清理1秒前的请求记录
            if ip in self.__requests:
                self.__requests[ip] = [
                    req_time
                    for req_time in self.__requests[ip]
                    if req_time > now - timedelta(seconds=1)
                ]
            else:
                self.__requests[ip] = []

            # 检查请求数量
            if len(self.__requests[ip]) >= count:
                return False

            # 记录本次请求
            self.__requests[ip].append(now)
            return True

    async def clear_ip_records(self, ip: str) -> None:
        """
        清除指定IP的所有记录
        """
        async with self.__limiter_lock:
            self.__requests.pop(ip, None)


class TokenManager:
    _instance = None
    __tokens_lock: Lock
    __tokens: dict[str, TokenData]
    __onetime_token_index: dict[str, OnetimeTokenData]  # one time token->token

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls.__tokens_lock = Lock()
            cls.__tokens = {}
            cls.__onetime_token_index = {}
        return cls._instance

    async def has_token(self, token: str) -> bool:
        async with self.__tokens_lock:
            return token in self.__tokens

    async def has_one_time_token(self, token: str) -> bool:
        async with self.__tokens_lock:
            if token in self.__onetime_token_index:
                auth: bool = False
                otk_data = self.__onetime_token_index[token]
                if not otk_data.expire < datetime.now(utc):
                    auth = True
                self.__onetime_token_index.pop(token, None)
                self.__tokens[otk_data.token].onetime_token = None
                return auth
            return False

    @overload
    async def get_token_data(self, token: str, /) -> TokenData: ...

    @overload
    async def get_token_data(self, token: str, default: T) -> TokenData | T: ...

    async def get_token_data(self, token: str, default: object = NOT_GIVEN):
        async with self.__tokens_lock:
            if default == NOT_GIVEN:
                return self.__tokens[token]
            else:
                return self.__tokens.get(token, default)

    @overload
    async def pop_token_data(self, token: str) -> TokenData: ...
    @overload
    async def pop_token_data(self, token: str, default: T) -> T | TokenData: ...

    async def pop_token_data(self, token: str, default: object = NOT_GIVEN):
        async with self.__tokens_lock:
            if default is not NOT_GIVEN:
                return self.__tokens.pop(token, default)
            else:
                return self.__tokens.pop(token)

    async def _expire_token_waiter(self, token: str, expire_at: datetime):
        await asyncio.sleep((expire_at - datetime.now(utc)).total_seconds())

        if data := await self.pop_token_data(token, None):
            async with self.__tokens_lock:
                if data.onetime_token:
                    self.__onetime_token_index.pop(data.onetime_token, None)

    async def _expire_onetime_token_waiter(self, token: str, expire_at: datetime):
        await asyncio.sleep((expire_at - datetime.now(utc)).total_seconds())
        if otk_data := self.__onetime_token_index.pop(token, None):
            self.__tokens[otk_data.token].onetime_token = None

    async def create_one_time_token(self, token_id: str) -> str:
        logger.debug(f"Creating one time token for '{token_id[:10]}...'")
        token = secrets.token_urlsafe(32)
        async with self.__tokens_lock:
            self.__tokens[token_id].onetime_token = token
            expire_at = datetime.now(utc) + timedelta(minutes=10)
            self.__onetime_token_index[token] = OnetimeTokenData(
                token=token_id, expire=expire_at
            )
            asyncio.create_task(self._expire_onetime_token_waiter(token, expire_at))  # noqa: RUF006
        return token

    async def create_access_token(
        self, data: dict, expires_delta: timedelta | None = None
    ):
        logger.debug(f"Creating access token for {data}")
        to_encode = data.copy()
        if expires_delta:
            expire = datetime.now(utc) + expires_delta
        else:
            expire = datetime.now(utc) + timedelta(minutes=30)
        to_encode.update({"exp": expire})
        encoded_jwt = secrets.token_urlsafe(32)

        async with self.__tokens_lock:
            self.__tokens[encoded_jwt] = TokenData(
                username=to_encode["sub"], expire=expire
            )
            asyncio.create_task(self._expire_token_waiter(encoded_jwt, expire))  # noqa: RUF006
        return encoded_jwt

    async def refresh_token(self, token: str) -> str:
        async with self.__tokens_lock:
            data_cache = self.__tokens[token]
            if onetime_token := data_cache.onetime_token:
                self.__onetime_token_index.pop(onetime_token, None)
            self.__tokens.pop(token, None)
        access_token_expires = timedelta(minutes=30)
        access_token = await self.create_access_token(
            data={"sub": data_cache.username}, expires_delta=access_token_expires
        )
        return access_token


class AuthManager:
    _instance = None
    _token_manager: TokenManager
    __users: dict[str, str]

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._token_manager = TokenManager()
            cls.__users = {
                get_webui_config().webui_user_name: cls._hash_password(
                    password=get_webui_config().webui_password
                )
            }
        return cls._instance

    async def check_otk_request(self, header: str):
        if not await self._token_manager.has_one_time_token(header):
            raise HTTPException(status_code=403, detail="访问不合法")

    async def check_current_user(self, request: Request):
        if otk := get_restful_auth_header(request):
            return await self.check_otk_request(otk)
        token = request.cookies.get(TOKEN_KEY)
        token_manager = self._token_manager
        if not token or (not await token_manager.has_token(token)):
            raise HTTPException(status_code=401, detail="未认证")
        token_data = await token_manager.get_token_data(token)
        if token_data.expire < datetime.now(utc):
            await token_manager.pop_token_data(token, None)
            raise HTTPException(status_code=401, detail="认证已过期")

    @staticmethod
    def _hash_password(*, password: str) -> str:
        salt = bcrypt.gensalt()
        hashed = bcrypt.hashpw(password.encode("utf-8"), salt)
        return hashed.decode("utf-8")

    def _verify_password(self, plain_password: str, hashed_password: str) -> bool:
        return bcrypt.checkpw(
            plain_password.encode("utf-8"), hashed_password.encode("utf-8")
        )

    async def authenticate_user(
        self, request: Request, username: str, password: str
    ) -> bool:
        # 检查速率限制
        client_ip = request.client.host if request.client else "unknown"
        rate_limiter = LoginRateLimiter()
        if not await rate_limiter.is_allowed(client_ip):
            logger.warning(f"Login rate limit exceeded for {client_ip}")
            raise HTTPException(
                status_code=403,
                detail="请求过于频繁，请稍后再试，重启Amrita以解除限制。",
            )

        if username in self.__users:
            result = self._verify_password(password, self.__users[username])
            # 如果认证成功，清除该IP的记录
            if result:
                await rate_limiter.clear_ip_records(client_ip)
            return result
        return False

    async def create_token(self, username: str, expire: timedelta) -> str:
        token = await self._token_manager.create_access_token(
            data={"sub": username}, expires_delta=expire
        )
        return token

    async def user_log_out(self, token: str):
        await self._token_manager.pop_token_data(token)

    async def refresh_token(self, request: Request):
        await self.check_current_user(request)
        token = request.cookies.get(TOKEN_KEY)
        if not token:
            raise HTTPException(status_code=401, detail="未认证")
        return await self._token_manager.refresh_token(token)



import sys
from collections.abc import Callable
from functools import wraps

import click

import amrita
from amrita.cli import check_optional_dependency, error


def require_full_depencies(func: Callable):
    def wrapper(*args, **kwargs):
        if not check_optional_dependency(quiet=True):
            click.echo(error("请使用 `uv add amrita[full]` 安装完整的可选依赖。"))
            sys.exit(1)
        return func(*args, **kwargs)

    return wrapper


def require_init(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs):
        amrita.init()
        amrita.load_plugins()
        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

import glob
import os
from pathlib import Path

import aiofiles
from fastapi import Request
from fastapi.responses import HTMLResponse, JSONResponse

from ..main import TemplatesManager, app
from ..sidebar import SideBarManager

# 获取项目根目录
PROJECT_ROOT = Path(os.getcwd())


@app.get("/bot/config", response_class=HTMLResponse)
async def config_editor(request: Request):
    """
    配置文件编辑器页面
    """
    # 获取所有.env文件
    env_files = glob.glob(str(PROJECT_ROOT / ".env*"))
    env_files = [
        Path(f).name for f in env_files if not f.endswith((".py", ".pyc", ".pyo"))
    ]

    # 默认选择.env文件
    selected_file = ".env"

    # 读取默认.env文件内容
    env_content = ""
    env_file_path = PROJECT_ROOT / selected_file
    if env_file_path.exists():
        async with aiofiles.open(env_file_path, encoding="utf-8") as f:
            env_content = await f.read()

    # 获取侧边栏
    side_bar = SideBarManager().get_sidebar_dump()
    for bar in side_bar:
        if bar.get("name") == "机器人管理":
            bar["active"] = True
            for child in bar.get("children", []):
                if child.get("name") == "配置管理":
                    child["active"] = True
                    break
            break

    return TemplatesManager().TemplateResponse(
        "config.html",
        {
            "request": request,
            "sidebar_items": side_bar,
            "env_content": env_content,
            "env_files": env_files,
            "selected_file": selected_file,
        },
    )


@app.post("/api/bot/config")
async def update_config(request: Request):
    """
    更新配置文件API
    """

    # 获取请求数据
    data = await request.json()
    content = data.get("content", "")
    filename = data.get("filename", ".env")

    if not content:
        return JSONResponse(
            status_code=400,
            content={"message": "Invalid request data"},
        )

    try:
        # 写入指定文件
        env_file_path = PROJECT_ROOT / filename
        async with aiofiles.open(env_file_path, "w", encoding="utf-8") as f:
            await f.write(content)
        return JSONResponse(
            {"code": 200, "message": f"配置文件 {filename} 更新成功", "error": None},
            200,
        )
    except Exception as e:
        return JSONResponse(
            {"code": 500, "message": "配置文件更新失败", "error": str(e)}, 500
        )


@app.get("/api/bot/config/{filename}")
async def get_config(filename: str):
    """
    获取指定配置文件内容
    """
    try:
        # 读取指定文件
        env_file_path = PROJECT_ROOT / filename
        if not env_file_path.exists():
            return JSONResponse(
                {"code": 404, "message": "文件不存在", "content": ""}, 404
            )

        async with aiofiles.open(env_file_path, encoding="utf-8") as f:
            content = await f.read()
        return JSONResponse(
            {"code": 200, "message": "success", "content": content}, 200
        )
    except Exception as e:
        return JSONResponse(
            {"code": 500, "message": "读取文件失败", "error": str(e)}, 500
        )



from nonebot.plugin import PluginMetadata, require

require("nonebot_plugin_localstore")
require("nonebot_plugin_orm")
require("amrita.plugins.menu")
from . import command_manager, config, on_init
from .commands import lp_chat_group, lp_perm_group, lp_user, main
from .config import DataManager

__all__ = [
    "DataManager",
    "command_manager",
    "config",
    "lp_chat_group",
    "lp_perm_group",
    "lp_user",
    "main",
    "on_init",
]

__plugin_meta__ = PluginMetadata(
    name="Amrita 权限管理模块",
    description="Amrita内置的权限组件",
    usage="https://amrita.suggar.top/amrita/plugins/liteperm/",
    homepage="https://github.com/AmritaBot/Amrita",
    type="library",
    supported_adapters={"~onebot.v11"},
)
